import npath from 'node:path'
import { assoc, concat, curry, curryN, lens, merge, pipe, prop, set, view } from 'rambda'
import { map, get as getMap } from './project.js'
import { get as getLinker } from './linker.js'
import { Cargo } from './cargo.js'
import { Crate } from './crates.js'

const optLevel = value => ({ 'opt-level': value })

//const assoc = curry((key, value, obj) => {
//    const l = lens(prop(key), assoc(key))
//    return set(l, value, obj)
//})

export const Bevy = Object.freeze({
    active: () => true,
    apply: curry(({ 'bevy-version': bevy }, project) =>
        map(
            'Cargo.toml',
            assoc('dependencies', { bevy }),
            project
        ))
})

export const DynamicLinking = Object.freeze({
    active: ({ dynlink }) => dynlink,
    apply: curry(({ 'bevy-version': version }, project) =>
        map(
            'Cargo.toml',
            assoc('dev-dependencies', { bevy: { version, features: ['dynamic_linking'] }}),
            project
        ))
})

export const DevelopmentOptimizations = Object.freeze({
    active: ({ dynopts }) => dynopts,
    apply: curry((_params, project) =>
        map(
            'Cargo.toml',
            pipe(
                assoc('profile.dev', optLevel(1)),
                assoc('profile.dev.package."*"', optLevel(3))
            )
        ))
})

export const ReleaseOptimizations = Object.freeze({
    active: ({ relopts }) => relopts,
    apply: curry((_params, project) => {
        const relOpts = { 'codegen-units': 1, lto: 'thin' }
        const wasmOpts = { inherits: 'release', 'opt-level': 's', strip: 'debuginfo' }

        return map(
            'Cargo.toml',
            pipe(
                assoc('profile.release', relOpts),
                assoc('profile.wasm-release', wasmOpts)
            ),
            project
        )
    })
})

export const CustomLinker = Object.freeze({
    active: ({ linker }) => linker,
    apply: curry(({ linker }, project) => {
        const linkinfo = getLinker(linker, project.host)

        if (!linkinfo) {
            return project
        }

        const { target, linker: bin, rustflags: rf1 = [] } = linkinfo
        const config = '.cargo/config.toml'
        const key = `target.${project.host}`
        const { rustflags: rf2 = [], ...rest } = getMap(config, project)[key] || {}

        return map(
            config,
            assoc(key, {
                ...rest,
                linker: bin,
                rustflags: concat(rf1, rf2)
            }),
            project
        )
    })
})

export const NightlyCompiler = Object.freeze({
    active: ({ nightly }) => nightly,
    apply: curry((_params, project) =>
        map(
            'rust-toolchain.toml',
            assoc('toolchain', { channel: 'nightly' }),
            project
        ))
})

export const Cranelift = Object.freeze({
    active: ({ cranelift, nightly }) => nightly && cranelift,
    apply: curry((_params, project) =>
        map(
            '.cargo/config.toml',
            pipe(
                assoc('unstable', { 'codegen-backend': true }),
                assoc('profile.dev', { 'codegen-backend': 'cranelift' }),
                assoc('profile.dev.package."*"', { 'codegen-backend': 'llvm' }),
            ),
            project
        ))
})

export const GenericSharing = Object.freeze({
    active: ({ genshare, nightly }) => nightly && genshare,
    apply: curry((_params, project) => {
        const config = '.cargo/config.toml'
        const key = `target.${project.host}`
        const { rustflags = [], ...rest } = getMap(config, project)[key] || {}

        return map(
            config,
            assoc(key, {
                ...rest,
                rustflags: [...rustflags, '-Zshare-generics=y'],
            }),
            project
        )
    })
})


export const ParallelFrontend = Object.freeze({
    active: ({ parallel, nightly }) => nightly && parallel,
    apply: curry((_params, project) => {
        const config = '.cargo/config.toml'
        const key = `target.${project.host}`
        const { rustflags = [], ...rest } = getMap(config, project)[key] || {}

        return map(
            config,
            assoc(key, {
                ...rest,
                rustflags: [...rustflags, '-Zthreads=0'],
            }),
            project
        )
    })
})
