import pkg from '../package.json' with { type: 'json' }
import parseArgs from 'minimist'
import path from 'node:path'
import { parse, stringify } from 'smol-toml'
import { reduce } from 'rambda'
import * as Project from './project.js'
import { Cargo } from './cargo.js'
import * as Configurations from './config.js'
import { Crate } from './crates.js'

const help = () => `Usage:
    ${pkg.name} PATH [options]

    Options:
        -b, --bevy-version  the bevy version to install, default to latest stable
        -l, --linker        configured linker. valid options: lld, mold (linux only)
        -q, --devopt        development optimizations
        -r, --relopt        release optimizations
        -d, --dynlink       enable dynamic linking
        -n, --nightly       enable nightly rust compiler
        -c, --cranelift     enable cranelift codegen (requires nightly)
        -g, --genshare      enable generic sharing (requires nightly)
        -p, --parallel      enable rustc's parallel frontend (requires nightly)

        -h, --help          print usage information and exit
        -v, --version       show version info and exit
`

const applicableConfigs = params =>
    Object.values(Configurations).filter(c => c.active(params))

const version = () => `${pkg.name} v${pkg.version}`

const main = async argv => {
    if(argv.help) {
        console.log(help())
    } else if (argv.version) {
        console.log(version())
    } else {
        const [relativeRoot] = argv._
        const root = path.join(process.cwd(), relativeRoot)
        await Cargo.init(root)
        const project = Project.create(path.normalize(root))
        const configs = applicableConfigs(argv)
        const result = await reduce(async (acc, cfg) => cfg.apply(argv, await acc), project, configs)
        await Project.serialize(result)
    }
}

const defaultBevyVersion = () =>
    Crate('bevy').then(c => c.json()).then(c => c.crate.default_version)

const defaultLinker = () => process.platform === 'darwin' ? undefined : 'lld'

const schema = Object.freeze({
    string: ['bevy-version', 'linker'],
    boolean: ['devopt', 'relopt', 'dynlink', 'nightly', 'cranelift', 'genshare', 'parallel', 'version'],
    alias: { b: 'bevy-version', l: 'linker', g: 'genshare', c: 'cranelift', n: 'nightly', d: 'dynlink', p: 'parallel', v: 'version', h: 'help' },
    default: { 'bevy-version': await defaultBevyVersion(), devopt: true, relopt: true, genshare: true, parallel: true, cranelift: true, nightly: true, dynlink: true, linker: defaultLinker() }
})

const argv = parseArgs(process.argv.slice(2), schema)

main(argv)

