import fs from 'node:fs/promises'
import npath from 'node:path'
import { parse, stringify } from 'smol-toml'
import { always, cond, curry, equals, flip, partial, pipe, T as True } from 'rambda'
import * as hamt from 'mini-hamt'

const guessHostTriplet = cond([
        [equals('win32'), always('x86_64-pc-windows-msvc')],
        [equals('darwin'), always('aarch64-apple-darwin')],
        [True, always('x86_64-unknown-linux-gnu')],
    ])

export const create = (root, host = guessHostTriplet(process.platform), modified = new Set(), config = hamt.empty) => Object.freeze({
    root,
    host,
    modified,
    config
})

const resolve = curry((path, { root }) => {
    return npath.isAbsolute(path) ? path : npath.join(root, path)
})

export const load = curry((path, project) => {
    const absPath = resolve(path, project)

    return fs.readFile(absPath, { encoding: 'utf-8' })
        .then(parse)
        .catch(() => ({}))
        .then(value => set(path, value, project))
})

export const get = curry((path, { config }) =>
    hamt.get(config, path)
)

export const set = curry((path, value, { root, host, modified, config }) => 
    create(
        root,
        host,
        new Set([path, ...modified]),
        hamt.set(config, path, value)
    )
)

export const map = curry(async (path, fn, project) => {
    let p = project
    let f = get(path, project)

    if (!f) {
        p = await load(path, project)
        f = get(path, p)
    }

    const r = fn(f)
    return set(path, r, p)
})

export const serialize = async project => {
    for(const path of Array.from(project.modified)) {
        const absPath = resolve(path, project)
        const dir = npath.dirname(absPath)
        await fs.mkdir(dir, { recursive: true })
        const data = Buffer.from(stringify(get(path, project)))
        fs.writeFile(absPath, data, { flag: 'w' })
    }
}

